# -*- coding: utf-8 -*-
"""Synteny (genome order) operations."""
# standard library imports
import sys

# from os.path import commonprefix as prefix
from pathlib import Path

# third-party imports
import attr
import click
import dask.bag as db
import numpy as np
import pandas as pd
from dask.diagnostics import ProgressBar

# first-party imports
import sh
import xxhash
from loguru import logger

# module imports
from . import cli
from . import click_loguru
from .common import ANCHOR_HIST_FILE
from .common import DIRECTIONAL_CATEGORY
from .common import HOMOLOGY_FILE
from .common import PROTEOMOLOGY_FILE
from .common import PROTEOSYN_FILE
from .common import SYNTENY_FILE
from .common import dotpath_to_path
from .common import read_tsv_or_parquet
from .common import write_tsv_or_parquet
from .mailboxes import DataMailboxes
from .mailboxes import ExternalMerge


@attr.s
class SyntenyBlockHasher(object):

    """Synteny-block hashes via reversible-peatmer method."""

    k = attr.ib(default=5)
    peatmer = attr.ib(default=True)
    prefix = attr.ib(default="tmp")

    def hash_name(self, no_prefix=False):
        """Return the string name of the hash function."""
        if no_prefix:
            prefix_str = ""
        else:
            prefix_str = self.prefix + "."
        if self.peatmer:
            return f"{prefix_str}hash.peatmer{self.k}"
        return f"{prefix_str}hash.kmer{self.k}"

    def _hash_kmer(self, kmer):
        """Return a hash of a kmer array."""
        return xxhash.xxh32_intdigest(kmer.tobytes())

    def shingle(self, cluster_series, base, direction, hash):
        """Return a vector of anchor ID's. """
        vec = cluster_series.to_numpy().astype(int)
        steps = np.insert((vec[1:] != vec[:-1]).astype(int), 0, 0).cumsum()
        try:
            assert max(steps) == self.k - 1
        except AssertionError:
            logger.error(
                f"Workaround for minor error in shingle at hash {hash}, base"
                f" {base};"
            )
            logger.error(f"input homology string={vec}")
            logger.error(f"max index = {max(steps)}, should be {self.k-1}")
            steps[np.where(steps > self.k - 1)] = self.k - 1
        if direction == "+":
            return base + steps
        return base + self.k - 1 - steps

    def calculate(self, cluster_series):
        """Return an array of synteny block hashes data."""
        # Maybe the best code I've ever written--JB
        vec = cluster_series.to_numpy().astype(int)
        if self.peatmer:
            uneq_idxs = np.append(np.where(vec[1:] != vec[:-1]), len(vec) - 1)
            runlengths = np.diff(np.append(-1, uneq_idxs))
            positions = np.cumsum(np.append(0, runlengths))[:-1]
            n_mers = len(positions) - self.k + 1
            footprints = pd.array(
                [runlengths[i : i + self.k].sum() for i in range(n_mers)],
                dtype=pd.UInt32Dtype(),
            )
        else:
            n_elements = len(cluster_series)
            n_mers = n_elements - self.k + 1
            positions = np.arange(n_elements)
            footprints = pd.array([self.k] * (n_mers), dtype=pd.UInt32Dtype())
        if n_mers < 1:
            return None
        # Calculate k-mers over indirect index
        kmer_mat = np.array(
            [vec[positions[i : i + self.k]] for i in range(n_mers)]
        )
        fwd_rev_hashes = np.array(
            [
                np.apply_along_axis(self._hash_kmer, 1, kmer_mat),
                np.apply_along_axis(
                    self._hash_kmer, 1, np.flip(kmer_mat, axis=1)
                ),
            ]
        )
        plus_minus = np.array([["+"] * n_mers, ["-"] * n_mers])
        directions = np.take_along_axis(
            plus_minus,
            np.expand_dims(fwd_rev_hashes.argmin(axis=0), axis=0),
            axis=0,
        )[0]
        return pd.DataFrame(
            [
                pd.Categorical(directions, dtype=DIRECTIONAL_CATEGORY),
                footprints,
                pd.array(
                    np.amin(fwd_rev_hashes, axis=0), dtype=pd.Int64Dtype()
                ),
            ],
            columns=["tmp.direction", "tmp.footprint", self.hash_name()],
            index=cluster_series.index[positions[:n_mers]],
        )


@cli.command()
@click_loguru.init_logger()
@click.option("-k", default=5, help="Synteny block length.", show_default=True)
@click.option(
    "--peatmer/--kmer",
    default=True,
    is_flag=True,
    show_default=True,
    help="Allow repeats in block.",
)
@click.option(
    "--parallel/--no-parallel",
    is_flag=True,
    default=True,
    show_default=True,
    help="Process in parallel.",
)
@click.argument("setname")
def synteny_anchors(k, peatmer, setname, parallel):
    """Calculate synteny anchors."""
    options = click_loguru.get_global_options()
    set_path = Path(setname)
    file_stats_path = set_path / PROTEOMOLOGY_FILE
    proteomes = read_tsv_or_parquet(file_stats_path)
    n_proteomes = len(proteomes)
    hasher = SyntenyBlockHasher(k=k, peatmer=peatmer)
    hash_mb = DataMailboxes(
        n_boxes=n_proteomes,
        mb_dir_path=(set_path / "mailboxes" / "hash_merge"),
    )
    hash_mb.write_headers("hash\n")
    arg_list = []
    n_hashes_list = []
    for idx, row in proteomes.iterrows():
        arg_list.append((idx, row["path"],))
    if parallel:
        bag = db.from_sequence(arg_list)
    if not options.quiet:
        logger.info(
            "Calculating synteny anchors using the"
            f" {hasher.hash_name(no_prefix=True)} function"
            + f" for {n_proteomes} proteomes"
        )
        ProgressBar().register()
    if parallel:
        n_hashes_list = bag.map(
            calculate_synteny_hashes, mailboxes=hash_mb, hasher=hasher
        ).compute()
    else:
        for args in arg_list:
            n_hashes_list.append(
                calculate_synteny_hashes(
                    args, mailboxes=hash_mb, hasher=hasher
                )
            )
    logger.info(f"Reducing {sum(n_hashes_list)} hashes via external merge")
    merger = ExternalMerge(
        file_path_func=hash_mb.path_to_mailbox, n_merge=n_proteomes
    )
    merger.init("hash")
    merged_hashes = merger.merge(
        count_key="tmp.ortho_count", ordinal_key="tmp.base"
    )
    merged_hashes["tmp.ortho_count"] *= k
    hash_mb.delete()
    ret_list = []
    if not options.quiet:
        logger.info(
            f"Merging {len(merged_hashes)} synteny anchors into {n_proteomes}"
            " proteomes"
        )
        ProgressBar().register()
    if parallel:
        ret_list = bag.map(
            merge_synteny_hashes, merged_hashes=merged_hashes, hasher=hasher,
        ).compute()
    else:
        for args in arg_list:
            ret_list.append(
                merge_synteny_hashes(
                    args, merged_hashes=merged_hashes, hasher=hasher
                )
            )
    synteny_stats = pd.DataFrame.from_dict(ret_list)
    synteny_stats = synteny_stats.set_index("idx").sort_index()
    with pd.option_context(
        "display.max_rows",
        None,
        "display.max_columns",
        None,
        "display.float_format",
        "{:,.2f}%".format,
    ):
        logger.info(synteny_stats)
    del synteny_stats["path"]
    proteomes = pd.concat([proteomes, synteny_stats], axis=1)
    write_tsv_or_parquet(proteomes, set_path / PROTEOSYN_FILE)


def calculate_synteny_hashes(args, mailboxes=None, hasher=None):
    """Calculate synteny hashes for protein genes."""
    idx, dotpath = args
    outpath = dotpath_to_path(dotpath)
    hom = read_tsv_or_parquet(outpath / HOMOLOGY_FILE)
    hom["tmp.nan_group"] = (
        (hom["hom.cluster"].isnull()).astype(int).cumsum() + 1
    ) * (~hom["hom.cluster"].isnull())
    hom.replace(to_replace={"tmp.nan_group": 0}, value=pd.NA, inplace=True)
    hash_name = hasher.hash_name()
    syn_list = []
    for unused_id_tuple, subframe in hom.groupby(
        by=["frag.id", "tmp.nan_group"]
    ):
        syn_list.append(hasher.calculate(subframe["hom.cluster"]))
    syn = pd.concat([df for df in syn_list if df is not None], axis=0)
    del syn_list
    syn["tmp.frag.id"] = syn.index.map(hom["frag.id"])
    syn["tmp.i"] = pd.array(range(len(syn)), dtype=pd.UInt32Dtype())
    hash_counts = syn[hash_name].value_counts()
    syn["tmp.self_count"] = pd.array(
        syn[hash_name].map(hash_counts), dtype=pd.UInt32Dtype()
    )
    frag_count_arr = pd.array([pd.NA] * len(syn), dtype=pd.UInt32Dtype())
    hash_is_null = syn[hash_name].isnull()
    for unused_frag, subframe in syn.groupby(by=["tmp.frag.id"]):
        try:
            frag_hash_counts = subframe[hash_name].value_counts()
        except ValueError:
            continue
        for unused_i, row in subframe.iterrows():
            row_no = row["tmp.i"]
            if not hash_is_null[row_no]:
                hash_val = row[hash_name]
                frag_count_arr[row_no] = frag_hash_counts[hash_val]
    syn["tmp.frag_count"] = frag_count_arr
    del syn["tmp.i"]
    write_tsv_or_parquet(syn, outpath / SYNTENY_FILE, remove_tmp=False)
    # Write out sorted list of hash values
    unique_hashes = syn[hash_name].unique().to_numpy()
    unique_hashes.sort()
    with mailboxes.locked_open_for_write(idx) as fh:
        np.savetxt(fh, unique_hashes, fmt="%d")
    return len(unique_hashes)


def merge_synteny_hashes(args, merged_hashes=None, hasher=None):
    """Merge synteny hashes into proteomes."""
    hash_name = hasher.hash_name()
    plain_hash_name = hasher.hash_name(no_prefix=True)
    idx, dotpath = args
    outpath = dotpath_to_path(dotpath)
    syn = read_tsv_or_parquet(outpath / SYNTENY_FILE)
    syn = syn.join(merged_hashes, on=hash_name)
    homology = read_tsv_or_parquet(outpath / HOMOLOGY_FILE)
    syn = pd.concat([homology, syn], axis=1)
    print(f"doing {dotpath}")
    n_prots = len(syn)
    syn["tmp.i"] = range(len(syn))
    shingled_vars = {
        plain_hash_name: np.array([np.nan] * n_prots),
        "direction": np.array([""] * n_prots),
        "self_count": np.array([np.nan] * n_prots),
        "footprint": np.array([np.nan] * n_prots),
        "frag_count": np.array([np.nan] * n_prots),
        "ortho_count": np.array([np.nan] * n_prots),
    }
    anchor_blocks = np.array([np.nan] * n_prots)
    for hash_val, subframe in syn.groupby(by=["tmp.base"]):
        for unused_i, row in subframe.iterrows():
            footprint = row["tmp.footprint"]
            row_no = row["tmp.i"]
            anchor_vec = hasher.shingle(
                syn["hom.cluster"][row_no : row_no + footprint],
                row["tmp.base"],
                row["tmp.direction"],
                row[hash_name],
            )
            anchor_blocks[row_no : row_no + footprint] = anchor_vec
            for key in shingled_vars:
                shingled_vars[key][row_no : row_no + footprint] = row[
                    "tmp." + key
                ]
    syn["syn.anchor_id"] = pd.array(anchor_blocks, dtype=pd.UInt32Dtype())
    syn["syn.direction"] = pd.Categorical(
        shingled_vars["direction"], dtype=DIRECTIONAL_CATEGORY
    )
    syn["syn." + plain_hash_name] = pd.array(
        shingled_vars[plain_hash_name], dtype=pd.Int64Dtype()
    )
    del shingled_vars["direction"], shingled_vars[plain_hash_name]
    for key in shingled_vars:
        syn["syn." + key] = pd.array(
            shingled_vars[key], dtype=pd.UInt32Dtype()
        )
    write_tsv_or_parquet(
        syn, outpath / SYNTENY_FILE,
    )
    in_synteny = n_prots - syn["syn.anchor_id"].isnull().sum()
    # n_assigned = n_genes - syn["hom.cluster"].isnull().sum()
    n_assigned = 1
    # Do histogram of blocks
    # anchor_hist = pd.DataFrame(anchor_counts.value_counts()).sort_index()
    # anchor_hist = anchor_hist.rename(columns={"syn.anchor_id": "self_count"})
    # anchor_hist["pct_anchors"] = anchor_hist["self_count"] * anchor_hist.index * 100.0 / n_assigned
    # write_tsv_or_parquet(anchor_hist, outpath / ANCHOR_HIST_FILE)
    ambig = (syn["tmp.self_count"] != 1).sum()
    synteny_pct = in_synteny * 100.0 / n_assigned
    unambig_pct = (in_synteny - ambig) * 100.0 / n_assigned
    synteny_stats = {
        "idx": idx,
        "path": dotpath,
        "hom.assign": n_assigned,
        "syn.anchors": in_synteny,
        "syn.ambig": ambig,
        "syn.assgn_pct": synteny_pct,
        "syn.unamb_pct": unambig_pct,
    }
    return synteny_stats


def dagchainer_id_to_int(ident):
    """Accept DAGchainer ids such as "cl1" and returns an integer."""
    if not ident.startswith("cl"):
        raise ValueError(f"Invalid ID {ident}.")
    id_val = ident[2:]
    if not id_val.isnumeric():
        raise ValueError(f"Non-numeric ID value in {ident}.")
    return int(id_val)


@cli.command()
@click_loguru.init_logger()
@click.argument("setname")
def dagchainer_synteny(setname):
    """Read DAGchainer synteny into homology frames.

    IDs must correspond between DAGchainer files and homology blocks.
    Currently does not calculate DAGchainer synteny.
    """

    cluster_path = Path.cwd() / "out_azulejo" / "clusters.tsv"
    if not cluster_path.exists():
        try:
            azulejo_tool = sh.Command("azulejo_tool")
        except sh.CommandNotFound:
            logger.error("azulejo_tool must be installed first.")
            sys.exit(1)
        logger.debug("Running azulejo_tool clean")
        try:
            output = azulejo_tool(["clean"])
        except sh.ErrorReturnCode:
            logger.error("Error in clean.")
            sys.exit(1)
        logger.debug("Running azulejo_tool run")
        try:
            output = azulejo_tool(["run"])
            print(output)
        except sh.ErrorReturnCode:
            logger.error(
                "Something went wrong in azulejo_tool, check installation."
            )
            sys.exit(1)
        if not cluster_path.exists():
            logger.error(
                "Something went wrong with DAGchainer run.  Please run it"
                " manually."
            )
            sys.exit(1)
    synteny_hash_name = "dagchainer"
    set_path = Path(setname)
    logger.debug(f"Reading {synteny_hash_name} synteny file.")
    syn = pd.read_csv(
        cluster_path, sep="\t", header=None, names=["hom.cluster", "id"]
    )
    syn["synteny_id"] = syn["hom.cluster"].map(dagchainer_id_to_int)
    syn = syn.drop(["hom.cluster"], axis=1)
    cluster_counts = syn["synteny_id"].value_counts()
    syn["synteny_count"] = syn["synteny_id"].map(cluster_counts)
    syn = syn.sort_values(by=["synteny_count", "synteny_id"])
    syn = syn.set_index(["id"])
    files_frame, frame_dict = read_files(setname)
    set_keys = list(files_frame["stem"])

    def id_to_synteny_property(ident, column):
        try:
            return int(syn.loc[ident, column])
        except KeyError:
            return 0

    for stem in set_keys:
        homology_frame = frame_dict[stem]
        homology_frame["synteny_id"] = homology_frame.index.map(
            lambda x: id_to_synteny_property(x, "synteny_id")
        )
        homology_frame["synteny_count"] = homology_frame.index.map(
            lambda x: id_to_synteny_property(x, "synteny_count")
        )
        synteny_name = f"{stem}-{synteny_hash_name}{SYNTENY_ENDING}"
        logger.debug(
            f"Writing {synteny_hash_name} synteny frame {synteny_name}."
        )
        homology_frame.to_csv(set_path / synteny_name, sep="\t")



# -*- coding: utf-8 -*-
"""Homology (sequence similarity) operations."""
# standard library imports
import fcntl
import json
import os
import shutil
from pathlib import Path

# third-party imports
import click
import dask.bag as db
import pandas as pd
from dask.diagnostics import ProgressBar

# first-party imports
import sh
from loguru import logger

# module imports
from . import cli
from . import click_loguru
from .common import CLUSTER_FILETYPE
from .common import CLUSTERS_FILE
from .common import FRAGMENTS_FILE
from .common import HOMOLOGY_FILE
from .common import PROTEINS_FILE
from .common import PROTEOMES_FILE
from .common import PROTEOMOLOGY_FILE
from .common import TrimmableMemoryMap
from .common import dotpath_to_path
from .common import group_key_filename
from .common import read_tsv_or_parquet
from .common import sort_proteome_frame
from .common import write_tsv_or_parquet
from .core import homology_cluster
from .mailboxes import DataMailboxes


@cli.command()
@click_loguru.init_logger()
@click.option(
    "--identity",
    "-i",
    default=0.0,
    help="Minimum sequence ID (0-1). [default: lowest]",
)
@click.option(
    "--parallel/--no-parallel",
    is_flag=True,
    default=True,
    show_default=True,
    help="Process in parallel.",
)
@click.argument("setname")
def do_homology(identity, setname, parallel):
    """Calculate homology clusters, MSAs, trees."""
    options = click_loguru.get_global_options()
    set_path = Path(setname)
    proteomes_path = set_path / PROTEOMES_FILE
    proteomes_in = read_tsv_or_parquet(proteomes_path)
    proteomes = sort_proteome_frame(proteomes_in)
    if not proteomes_in.equals(proteomes):
        logger.info("proteomes sort order changed, writing new proteomes file")
        write_tsv_or_parquet(proteomes, proteomes_path)
    n_proteomes = len(proteomes)
    frags = read_tsv_or_parquet(set_path / FRAGMENTS_FILE)
    frag_frames = {}
    for dotpath, subframe in frags.groupby(by=["path"]):
        frag_frames[dotpath] = subframe.copy().set_index("frag.orig_id")
    # write concatenated stats
    concat_fasta_path = set_path / "proteins.fa"
    if concat_fasta_path.exists():
        concat_fasta_path.unlink()
    arg_list = []
    for i, row in proteomes.iterrows():
        arg_list.append((row, concat_fasta_path, frag_frames[row["path"]]))
    if not options.quiet:
        logger.info(f"Concatenating sequences for {len(arg_list)} proteomes:")
    for args in arg_list:
        write_concatenated_protein_fasta(args)
    del arg_list
    file_idx = {}
    stem_dict = {}
    for i, row in proteomes.iterrows():
        stem = row["path"]
        file_idx[stem] = i
        stem_dict[i] = stem
    logger.debug("Doing cluster calculation.")
    cwd = Path.cwd()
    os.chdir(set_path)
    n_clusters, run_stats, cluster_hist = homology_cluster.callback(
        "proteins.fa",
        identity,
        write_ids=True,
        delete=False,
        cluster_stats=False,
        outname="homology",
    )
    log_path = Path("homology.log")
    shutil.copy2(log_path, "logs/homology.log")
    log_path.unlink()
    os.chdir(cwd)
    # logger.info(f"Stats of {n_clusters} clusters:")
    # logger.info(run_stats)
    # logger.info(f"\nCluster size histogram ({n_proteomes} proteomes):")
    # with pd.option_context(
    #     "display.max_rows", None, "display.float_format", "{:,.2f}%".format
    # ):
    #    logger.info(cluster_hist)
    del cluster_hist
    del run_stats
    concat_fasta_path.unlink()
    mb = DataMailboxes(
        n_boxes=n_proteomes,
        mb_dir_path=(set_path / "mailboxes" / "clusters2proteomes"),
        file_extension="tsv",
    )
    mb.write_headers("\thom.cluster\thom.siz\n")
    cluster_paths = [
        set_path / "homology" / f"{i}.fa" for i in range(n_clusters)
    ]
    if parallel:
        bag = db.from_sequence(cluster_paths)
    else:
        cluster_stats = []
    if not options.quiet:
        logger.info(
            f"Calculating MSAs and trees for {len(cluster_paths)} homology"
            " clusters:"
        )
        ProgressBar().register()
    if parallel:
        cluster_stats = bag.map(
            parse_cluster,
            file_dict=file_idx,
            file_writer=mb.locked_open_for_write,
        )
    else:
        for clust_fasta in cluster_paths:
            cluster_stats.append(
                parse_cluster(
                    clust_fasta,
                    file_dict=file_idx,
                    file_writer=mb.locked_open_for_write,
                )
            )
    n_clust_genes = 0
    clusters_dict = {}
    for cluster_id, cluster_dict in cluster_stats:
        n_clust_genes += cluster_dict["size"]
        clusters_dict[cluster_id] = cluster_dict
    del cluster_stats
    clusters = pd.DataFrame.from_dict(clusters_dict).transpose()
    del clusters_dict
    clusters.sort_index(inplace=True)
    grouping_dict = {}
    for i in range(n_proteomes):  # keep numbering of single-file clusters
        grouping_dict[f"[{i}]"] = i
    grouping_dict[str(list(range(n_proteomes)))] = 0
    for n_members, subframe in clusters.groupby(["n_memb"]):
        if n_members == 1:
            continue
        if n_members == n_proteomes:
            continue
        member_counts = pd.DataFrame(subframe["members"].value_counts())
        member_counts["key"] = range(len(member_counts))
        for newcol in range(n_members):
            member_counts[f"memb{newcol}"] = ""
        for member_string, row in member_counts.iterrows():
            grouping_dict[member_string] = row["key"]
            member_list = json.loads(member_string)
            for col in range(n_members):
                member_counts.loc[member_string, f"memb{col}"] = stem_dict[
                    member_list[col]
                ]
        member_counts = member_counts.set_index("key")
        write_tsv_or_parquet(
            member_counts, set_path / group_key_filename(n_members)
        )
    clusters["members"] = clusters["members"].map(grouping_dict)
    clusters = clusters.rename(columns={"members": "group_key"})
    n_adj = clusters["n_adj"].sum()
    adj_pct = n_adj * 100.0 / n_clust_genes
    n_adj_clust = sum(clusters["adj_groups"] != 0)
    adj_clust_pct = n_adj_clust * 100.0 / len(clusters)
    logger.info(
        f"{n_adj} ({adj_pct:.1f}%) out of {n_clust_genes}"
        + " clustered genes are adjacent"
    )
    logger.info(
        f"{n_adj_clust} ({adj_clust_pct:.1f}%) out of "
        + f"{len(clusters)} clusters contain adjacency"
    )
    write_tsv_or_parquet(clusters, set_path / CLUSTERS_FILE)
    # join homology cluster info to proteome info
    arg_list = []
    for i, row in proteomes.iterrows():
        arg_list.append((i, dotpath_to_path(row["path"]),))
    if parallel:
        bag = db.from_sequence(arg_list)
    else:
        homo_stats = []
    if not options.quiet:
        logger.info(f"Joining homology info to {n_proteomes} proteomes:")
        ProgressBar().register()
    if parallel:
        homo_stats = bag.map(
            join_homology_to_proteome, mailbox_reader=mb.open_then_delete
        ).compute()
    else:
        for args in arg_list:
            homo_stats.append(
                join_homology_to_proteome(
                    args, mailbox_reader=mb.open_then_delete
                )
            )
    mb.delete()
    homo_frame = pd.DataFrame.from_dict(homo_stats)
    homo_frame.set_index(["idx"], inplace=True)
    homo_frame.sort_index(inplace=True)
    logger.info("Homology cluster coverage:")
    with pd.option_context(
        "display.max_rows", None, "display.float_format", "{:,.2f}%".format
    ):
        logger.info(homo_frame)
    proteomes = pd.concat([proteomes, homo_frame], axis=1)
    write_tsv_or_parquet(
        proteomes, set_path / PROTEOMOLOGY_FILE, float_format="%5.2f"
    )


def write_concatenated_protein_fasta(args):
    row, concat_fasta_path, frags = args
    """Read peptide sequences from info file and write them out."""
    dotpath = row["path"]
    phylogeny_dict = {"idx": row.name, "path": dotpath}
    for n in [name for name in row.index if name.startswith("phy.")]:
        phylogeny_dict[n] = row[n]
    inpath = dotpath_to_path(dotpath) / PROTEINS_FILE
    prot_info = pd.read_parquet(inpath)
    prot_info["frag.is_plas"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.is_plas"]
    )
    prot_info["frag.is_scaf"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.is_scaf"]
    )
    prot_info["frag.is_chr"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.is_chr"]
    )
    prot_info["frag.id"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.id"]
    )
    for prop in phylogeny_dict:
        prot_info[prop] = phylogeny_dict[prop]
    info_to_fasta.callback(
        None, concat_fasta_path, append=True, infoobj=prot_info
    )


def parse_cluster(
    fasta_path, file_dict=None, file_writer=None, neighbor_joining=False
):
    """Parse cluster FASTA headers to create cluster table.."""
    cluster_id = fasta_path.name[:-3]
    outdir = fasta_path.parent
    clusters = parse_cluster_fasta(fasta_path)
    if len(clusters) < 2:
        fasta_path.unlink()
        raise ValueError("Singleton Cluster")
    # calculate MSA and return guide tree
    muscle_args = [
        "-in",
        f"{outdir}/{cluster_id}.fa",
        "-out",
        f"{outdir}/{cluster_id}.faa",
        "-diags",
        "-sv",
        "-maxiters",
        "2",
        "-quiet",
        "-distance1",
        "kmer20_4",
    ]
    if len(clusters) >= 4:
        muscle_args += [
            "-tree2",
            f"{outdir}/{cluster_id}.nwk",
        ]
        if neighbor_joining:
            muscle_args += ["-cluster2", "neighborjoining"]  # adds 20%
    muscle = sh.Command("muscle")
    muscle(muscle_args)
    fasta_path.unlink()
    clusters["idx"] = clusters["path"].map(file_dict)
    clusters.sort_values(by=["idx", "frag.id", "frag.pos"], inplace=True)
    clusters["adj_group"] = ""
    adjacency_group = 0
    was_adj = False
    for unused_group_id, subframe in clusters.groupby(by=["idx", "frag.id"]):
        if len(subframe) == 1:
            continue
        last_pos = -2
        last_ID = None
        if was_adj:
            adjacency_group += 1
        was_adj = False
        for gene_id, row in subframe.iterrows():
            if row["frag.pos"] == last_pos + 1:
                if not was_adj:
                    clusters.loc[last_ID, "adj_group"] = str(adjacency_group)
                was_adj = True
                clusters.loc[gene_id, "adj_group"] = str(adjacency_group)
            else:
                if was_adj:
                    adjacency_group += 1
                    was_adj = False
            last_pos = row["frag.pos"]
            last_ID = gene_id
    if was_adj:
        adjacency_group += 1
    idx_values = clusters["idx"].value_counts()
    idx_list = list(idx_values.index)
    idx_list.sort()
    write_tsv_or_parquet(clusters, outdir / f"{cluster_id}.{CLUSTER_FILETYPE}")
    n_adj = sum(clusters["adj_group"] != "")
    cluster_dict = {
        "size": len(clusters),
        "n_memb": len(idx_values),
        "members": str(idx_list),
        "n_adj": n_adj,
        "adj_groups": adjacency_group,
    }
    for group_id, subframe in clusters.groupby(by=["idx"]):
        proteome_frame = subframe.copy()
        proteome_frame["hom.cluster"] = cluster_id
        proteome_frame["hom.cl_size"] = len(idx_values)
        proteome_frame.drop(
            proteome_frame.columns.drop(
                ["hom.cluster", "hom.cl_size"]
            ),  # drop EXCEPT these
            axis=1,
            inplace=True,
        )
        with file_writer(group_id) as fh:
            proteome_frame.to_csv(fh, header=False, sep="\t")
    return int(cluster_id), cluster_dict


def parse_cluster_fasta(filepath, trim_dict=True):
    """Return FASTA headers as a dictionary of properties."""
    next_pos = 0
    properties_dict = {}
    memory_map = TrimmableMemoryMap(filepath)
    with memory_map.map() as mm:
        size = memory_map.size
        next_pos = mm.find(b">", next_pos)
        while next_pos != -1 and next_pos < size:
            eol_pos = mm.find(b"\n", next_pos)
            if eol_pos == -1:
                break
            space_pos = mm.find(b" ", next_pos + 1, eol_pos)
            if space_pos == -1:
                raise ValueError(
                    f"Header format is bad in {filepath} header"
                    f" {len(properties_dict)+1}"
                )
            id = mm[next_pos + 1 : space_pos].decode("utf-8")
            payload = json.loads(mm[space_pos + 1 : eol_pos])
            properties_dict[id] = payload
            if trim_dict:
                size = memory_map.trim(space_pos, eol_pos)
            next_pos = mm.find(b">", space_pos)
    cluster = (
        pd.DataFrame.from_dict(properties_dict).transpose().convert_dtypes()
    )
    return cluster


def join_homology_to_proteome(args, mailbox_reader=None):
    """Read homology info from mailbox and join it to proteome file."""
    idx, protein_parent = args
    proteins = pd.read_parquet(protein_parent / PROTEINS_FILE)
    n_proteins = len(proteins)
    with mailbox_reader(idx) as fh:
        homology_frame = pd.read_csv(
            fh, sep="\t", index_col=0
        ).convert_dtypes()
        clusters_in_proteome = len(homology_frame)
    proteome_frame = pd.concat([proteins, homology_frame], axis=1)
    write_tsv_or_parquet(proteome_frame, protein_parent / HOMOLOGY_FILE)
    return {
        "idx": idx,
        "clustered": clusters_in_proteome,
        "cluster_pct": clusters_in_proteome * 100.0 / n_proteins,
    }


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--append/--no-append",
    "-a/-x",
    is_flag=True,
    default=True,
    help="Append to FASTA file.",
    show_default=True,
)
@click.argument("infofile")
@click.argument("fastafile")
def info_to_fasta(infofile, fastafile, append, infoobj=None):
    """Convert infofile to FASTA file."""
    if infoobj is None:
        infoobj = read_tsv_or_parquet(infofile)
    if append:
        filemode = "a+"
    else:
        filemode = "w"
    with Path(fastafile).open(filemode) as fh:
        fcntl.flock(fh, fcntl.LOCK_EX)
        logger.debug(f"Writing to {fastafile} with mode {filemode}.")
        seqs = infoobj["prot.seq"].copy()
        del infoobj["prot.seq"]
        for gene_id, row in infoobj.iterrows():
            fh.write(f">{gene_id} {row.to_json()}\n")
            fh.write(f"{seqs[gene_id]}\n")
        fcntl.flock(fh, fcntl.LOCK_UN)



# -*- coding: utf-8 -*-
"""Synteny (genome order) operations."""
# standard library imports
import statistics
import sys

# from os.path import commonprefix as prefix
from pathlib import Path

# third-party imports
import click
import numpy as np
import pandas as pd

# first-party imports
from loguru import logger

# module imports
from . import cli
from . import click_loguru


def read_files(setname, synteny=None):
    """Read previously-calculated homology/synteny files and file frame."""
    set_path = Path(setname)
    files_frame_path = set_path / f"{setname}{FILES_ENDING}"
    try:
        file_frame = pd.read_csv(files_frame_path, index_col=0, sep="\t")
    except FileNotFoundError:
        logger.error(f"Unable to read files frome from {files_frame_path}")
        sys.exit(1)
    if synteny is None:
        ending = HOMOLOGY_FILE
        file_type = "homology"
    else:
        ending = f"-{synteny}-{SYNTENY_ENDING}"
        file_type = "synteny"
    paths = list(set_path.glob("*" + ending))
    stems = [p.name[: -len(ending)] for p in paths]
    if len(stems) != len(file_frame):
        logger.error(
            f"Number of {file_type} files ({len(stems)})"
            + f"is not the same as length of file frame({len(file_frame)})."
        )
        sys.exit(1)
    frame_dict = {}
    for i, path in enumerate(paths):
        logger.debug(f"Reading homology file {path}.")
        frame_dict[stems[i]] = pd.read_csv(path, index_col=0, sep="\t")
    return file_frame, frame_dict


class ProxySelector:

    """Provide methods for downselection of proxy genes."""

    def __init__(self, frame, prefs):
        """Calculate any joint statistics from frame."""
        self.frame = frame
        self.prefs = prefs
        self.reasons = []
        self.drop_ids = []
        self.first_choice = prefs[0]
        self.first_choice_hits = 0
        self.first_choice_unavailable = 0
        self.cluster_count = 0

    def choose(self, chosen_one, cluster, reason, drop_non_chosen=True):
        """Make the choice, recording stats."""
        self.frame.loc[chosen_one, "reason"] = reason
        self.first_choice_unavailable += int(
            self.first_choice not in set(cluster["stem"])
        )
        self.first_choice_hits += int(
            cluster.loc[chosen_one, "stem"] == self.first_choice
        )
        non_chosen_ones = list(cluster.index)
        non_chosen_ones.remove(chosen_one)
        if drop_non_chosen:
            self.drop_ids += non_chosen_ones
        else:
            self.cluster_count += len(non_chosen_ones)

    def choose_by_preference(
        self, subcluster, cluster, reason, drop_non_chosen=True
    ):
        """Choose in order of preference."""
        stems = subcluster["stem"]
        pref_idxs = [subcluster[stems == pref].index for pref in self.prefs]
        pref_lens = np.array([int(len(idx) > 0) for idx in pref_idxs])
        best_choice = np.argmax(pref_lens)  # first occurrance
        if pref_lens[best_choice] > 1:
            raise ValueError(
                f"subcluster {subcluster} is not unique w.r.t. proteome"
                f" {list(stems)[best_choice]}."
            )
        self.choose(
            pref_idxs[best_choice][0], cluster, reason, drop_non_chosen
        )

    def choose_by_length(self, subcluster, cluster, drop_non_chosen=True):
        """Return an index corresponding to the selected modal/median length."""
        counts = subcluster["protein_len"].value_counts()
        max_count = max(counts)
        if max_count > 1:  # repeated values exist
            max_vals = list(counts[counts == max(counts)].index)
            modal_cluster = subcluster[
                subcluster["protein_len"].isin(max_vals)
            ]
            self.choose_by_preference(
                modal_cluster,
                cluster,
                f"mode{len(modal_cluster)}",
                drop_non_chosen=drop_non_chosen,
            )
        else:
            lengths = list(subcluster["protein_len"])
            median_vals = [
                statistics.median_low(lengths),
                statistics.median_high(lengths),
            ]
            median_pair = subcluster[
                subcluster["protein_len"].isin(median_vals)
            ]
            self.choose_by_preference(
                median_pair, cluster, "median", drop_non_chosen=drop_non_chosen
            )

    def cluster_selector(self, cluster):
        """Calculate which gene in a homology cluster should be left and why."""
        self.cluster_count += 1
        if len(cluster) == 1:
            self.choose(cluster.index[0], cluster, "singleton")
        else:
            for synteny_id, subcluster in cluster.groupby(by=["synteny_id"]):
                if len(subcluster) > 1:
                    self.choose_by_length(
                        subcluster, cluster, drop_non_chosen=(not synteny_id)
                    )
                else:
                    if subcluster["synteny_id"][0] != 0:
                        self.choose(
                            subcluster.index[0],
                            cluster,
                            "bad_synteny",
                            drop_non_chosen=(not synteny_id),
                        )
                    else:
                        self.choose(
                            subcluster.index[0],
                            cluster,
                            "single",
                            drop_non_chosen=(not synteny_id),
                        )

    def downselect_frame(self):
        """Return a frame with reasons for keeping and non-chosen-ones dropped."""
        drop_pct = len(self.drop_ids) * 100.0 / len(self.frame)
        logger.info(
            f"Dropping {len(self.drop_ids)} ({drop_pct:0.1f}%) of"
            f" {len(self.frame)} genes."
        )
        return self.frame.drop(self.drop_ids)

    def selection_stats(self):
        """Return selection stats."""
        return (
            self.cluster_count,
            self.first_choice_unavailable,
            self.first_choice_hits,
        )


@cli.command()
@click_loguru.init_logger()
@click.argument("setname")
@click.argument("synteny_type")
@click.argument("prefs", nargs=-1)
def calculate_proxy_genes(setname, synteny_type, prefs):
    """Calculate a set of proxy genes from synteny files.

    prefs is an optional list of proteome stems in order of preference in the proxy calc.
    """
    set_path = Path(setname)
    files_frame, frame_dict = read_files(setname, synteny=synteny_type)
    set_keys = list(files_frame["stem"])
    default_prefs = set_keys.copy()
    default_prefs.reverse()
    if prefs != ():
        for stem in prefs:
            if stem not in default_prefs:
                logger.error(f"Preference {stem} not in {default_prefs}")
                sys.exit(1)
            else:
                default_prefs.remove(stem)
        prefs = list(prefs) + default_prefs
        order = "non-default"
    else:
        prefs = default_prefs
        order = "default"
    logger.debug(
        f"Genome preference for proxy selection in {order} order: {prefs}"
    )
    proxy_frame = None
    for stem in set_keys:
        logger.debug(f"Reading {stem}")
        frame_dict[stem]["stem"] = stem
        if proxy_frame is None:
            proxy_frame = frame_dict[stem]
        else:
            proxy_frame = proxy_frame.append(frame_dict[stem])
    del files_frame
    proxy_frame = proxy_frame.sort_values(
        by=["cluster_size", "hom.cluster", "synteny_count", "synteny_id"]
    )
    proxy_filename = f"{setname}-{synteny_type}-{PROXY_ENDING}"
    logger.debug(f"Writing initial proxy file {proxy_filename}.")
    proxy_frame.to_csv(set_path / proxy_filename, sep="\t")
    proxy_frame["reason"] = ""
    logger.debug("Downselecting homology clusters.")
    downselector = ProxySelector(proxy_frame, prefs)
    for unused_cluster, homology_cluster in proxy_frame.groupby(
        by=["hom.cluster"]
    ):  # pylint: disable=unused-variable
        downselector.cluster_selector(homology_cluster)
    downselected = downselector.downselect_frame()
    downselected_filename = (
        f"{setname}-{synteny_type}-downselected-{PROXY_ENDING}"
    )
    logger.debug(f"Writing downselected proxy file {downselected_filename}.")
    downselected.to_csv(set_path / downselected_filename, sep="\t")
    # print out stats
    (
        cluster_count,
        first_choice_unavailable,
        first_choice_hits,
    ) = downselector.selection_stats()
    first_choice_percent = (
        first_choice_hits * 100.0 / (cluster_count - first_choice_unavailable)
    )
    first_choice_unavailable_percent = (
        first_choice_unavailable * 100.0 / cluster_count
    )
    logger.info(
        f"First-choice ({prefs[0]}) selections from {cluster_count} homology"
        " clusters:"
    )
    logger.info(
        f"   not in cluster: {first_choice_unavailable}"
        f" ({first_choice_unavailable_percent:.1f}%)"
    )
    logger.info(
        f"   chosen as proxy: {first_choice_hits}"
        f" ({first_choice_percent:.1f}%)"
    )



# -*- coding: utf-8 -*-
"""Data analysis and plotting."""
# standard library imports
import sys
from pathlib import Path

# third-party imports
import click
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

# first-party imports
from loguru import logger

# module imports
from . import cli
from . import click_loguru
from .common import cluster_set_name
from .common import homo_degree_dist_filename

# Global constants
NAME = "azulejo"
STATFILE_SUFFIX = f"-{NAME}_stats.tsv"
ANYFILE_SUFFIX = f"-{NAME}_ids-any.tsv"
ALLFILE_SUFFIX = f"-{NAME}_ids-all.tsv"
IDENT_LOG_MIN = -3
IDENT_LOG_MAX = 0
EPSILON = 0.000001
FILETYPE = "png"
MAX_BINS = 10


def make_histogram(dist, name, log10=False):
    """Do histogram plot with kernel density estimate."""
    dist = dist[dist < 10]
    mean = dist.mean()
    if log10:
        dist = np.log10(dist)
    # if len(dist) < MAX_BINS:
    #    bins = len(dist)
    # else:
    #    bins = MAX_BINS
    sns.distplot(
        dist,
        bins=None,
        rug=False,
        kde=False,
        norm_hist=True,
        rug_kws={"color": "b"},
        kde_kws={"color": "k", "linewidth": 1, "label": "KDE"},
        hist_kws={
            "histtype": "step",
            "linewidth": 2,
            "alpha": 1,
            "color": "b",
            #'range':(0,20)
        },
    )
    plt.title(f"{name} histogram of {len(dist):d} values, mean={mean:.1f}")
    if log10:
        plt.xlabel("log " + name)
    else:
        plt.xlabel(name)
    plt.ylabel("Frequency")
    # for ext in PLOT_TYPES:
    #    plt.savefig(LOG_PATH / ('%s-histogram.' % (name.rstrip('%')) + ext),
    #                bbox_inches='tight')
    plt.show()
    # plt.close('all')


def tick_function(tick_x):
    """Compute ticks."""
    tick_x = tick_x * 3.0 - 3
    vals = [
        (f"{v:f}").rstrip("0").rstrip(".")
        for v in (1.0 - 10 ** tick_x) * 100.0
    ]
    ticks = [f"{v}%" for v in vals]
    return ticks


def log_deriv(xvals, yvals):
    """Compute the logarithmic derivative."""
    log_x = -1.0 * np.log10(xvals + EPSILON)
    log_y = np.log10(yvals)
    return np.gradient(log_y) / np.gradient(log_x)


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("instemlist")
def analyze_clusters(
    dirname, instemlist, label, reference=None, on_id=None, match_type=None
):
    """Statistics of clustering as function of identity."""
    if match_type is None:
        matches = ["all", "any"]
    else:
        matches = [match_type]
    uniques = {}
    divergence = {}
    dirpath = Path(dirname)
    div_dist = {"all": {"ref": 0.0}, "any": {"ref": 0.0}}
    print("ref=", reference)
    for stem in instemlist:
        print("stem=", stem)
        paths = {
            "all": dirpath / (stem + ALLFILE_SUFFIX),
            "any": dirpath / (stem + ANYFILE_SUFFIX),
            "stat": dirpath / (stem + STATFILE_SUFFIX),
        }
        stats = pd.read_csv(paths["stat"], sep="\t", index_col=0)
        uniques[stem] = stats["unique_seqs"].iloc[0]
        divergence[stem] = stats["divergence"]
        if on_id is None:
            div_dist["all"][stem] = log_deriv(
                divergence[stem], stats["clusters"]
            )
            div_dist["any"][stem] = None
            if stem == reference:
                div_dist["all"]["ref"] = div_dist["all"][stem]
                div_dist["any"]["ref"] = None
        else:
            for match in ["any", "all"]:
                data = pd.read_csv(paths[match], sep="\t", index_col=0)
                try:
                    div_dist[match][stem] = log_deriv(
                        divergence[stem], data.loc[on_id]
                    )
                except KeyError:  # this label wasn't found
                    div_dist[match][stem] = None
                if stem == reference:
                    div_dist[match]["ref"] = div_dist[match][stem]
    #
    # Make the plots
    #
    plt.style.use("seaborn-whitegrid")
    axis_dict = {}
    fig, axes = plt.subplots(len(matches), sharex=True)
    try:
        for axis, i in enumerate(axes):
            axis_dict[matches[i]] = axis
            loweraxis = axes[1]
    except TypeError:
        axis_dict[matches[0]] = axes
        loweraxis = axes
    for stem in instemlist:
        for match in matches:
            if div_dist[match][stem] is None:
                continue
            axis_dict[match].plot(
                divergence[stem],
                div_dist[match][stem] - div_dist[match]["ref"],
                label=f"{stem.replace(label + '.', '')}",
            )
            # uniques[stem]/1000.))
    if reference is None:
        if on_id is None:
            title = f"{label} Divergence Distribution"
            outfilestem = f"{label}_divergence_dist."
        else:
            title = f'{label} Divergence Distribution on "{on_id}"'
            outfilestem = f"{label}_divergence_dist_{on_id}."
    else:
        if on_id is None:
            title = (
                f"{label}_Differential Divergence Distribution vs. {reference}"
            )
            outfilestem = f"{label}_divergence_dist_vs{reference}."
        else:
            title = (
                f'{label} Differential Divergence Distribution on "{on_id}"'
                f" vs. {reference}"
            )
            outfilestem = f"{label}_divergence_dist_on_{on_id}_vs_ref."
    if reference is None:
        fig.text(
            0.02,
            0.5,
            "Logarithmic Derivative on Clusters",
            ha="center",
            va="center",
            rotation="vertical",
        )
    else:
        fig.text(
            0.02,
            0.5,
            "Logarithmic Derivative Difference on Clusters",
            ha="center",
            va="center",
            rotation="vertical",
        )
    if len(matches) == 2:
        fig.text(0.5, 0.47, "All in Cluster", ha="center", va="center")
        fig.text(0.5, 0.89, "Any in Cluster", ha="center", va="center")
    else:
        fig.text(
            0.5,
            0.91,
            f"{matches[0].capitalize()} in Cluster",
            ha="center",
            va="center",
        )
    loweraxis.set(xlabel="Divergence on Sequence Identity")
    loweraxis.legend(loc="upper left")
    fig.suptitle(title)
    plt.xscale("log")
    limits = [0.001, 1.0]
    new_tick_locations = np.array([0.0, 1.0 / 3.0, 2.0 / 3.0, 1.0])
    # loweratick_xis.set_xlim(limits)
    axis_dict["second"] = loweraxis.twiny()
    axis_dict["second"].set_xlim(limits)
    axis_dict["second"].set_xticks(new_tick_locations)
    axis_dict["second"].set_xticklabels(tick_function(new_tick_locations))
    axis_dict["second"].set_xlabel("   ")
    # r'%Identity')
    # plt.ylim([-0.002,0.002])
    outfilename = outfilestem + f"{FILETYPE}"
    print(f"saving plot to {outfilename}")
    plt.savefig(dirpath / outfilename, dpi=200)
    plt.show()


def do_cuts(obs, high, low, label):
    """Cut at high and low levels."""
    if high > 0.0:
        hicuts = obs[obs > high]
        obs = obs[obs <= high]
        if len(hicuts) == 0:
            hifilename = label + "_hicuts.tsv"
            logger.info(
                "%d observations dropped by high-side cutoff of %.2f written"
                " to %s",
                len(hicuts),
                high,
                hifilename,
            )
            logger.info(hicuts)
    if low > 0.0:
        locuts = obs[obs > low]
        obs = obs[obs >= low]
        logger.info(
            "%d observations dropped by low-side cutoff of %.2f",
            len(locuts),
            low,
        )
        if len(locuts) == 0:
            lofilename = label + "_locuts.tsv"
            logger.info(
                "%d observations dropped by low-side cutoff of %.2f written"
                " to %s",
                len(locuts),
                low,
                lofilename,
            )
            logger.info(locuts)
    return obs


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--hi_cutoff",
    default=2.0,
    show_default=True,
    help="Disregard above this value.",
)
@click.option(
    "--lo_cutoff",
    default=0.0,
    show_default=True,
    help="Disregard below this value.",
)
@click.argument("cluster_size")
@click.argument("combinedfile")
def outlier_length_dist(hi_cutoff, lo_cutoff, cluster_size, combinedfile):
    """Plot length distribution of outliers in clusters."""
    cluster_size = int(cluster_size)
    if cluster_size <= 0:
        logger.error("Positive cluster size must be specified")
        sys.exit(1)
    clusters = pd.read_csv(combinedfile, sep="\t", index_col=0)
    norm_lengths = []
    for unused_cluster_id, cluster in clusters.groupby(
        "hom.cluster"
    ):  # pylint: disable=unused-variable
        if cluster["siz"].iloc[0] != cluster_size:
            # not the right size
            continue
        if len(set(cluster["sub"])) != 2:
            # not just two subclusters
            continue
        if 1 not in set(cluster["sub_siz"]):
            # no singleton cluster
            continue
        singleton = cluster[cluster["sub_siz"] == 1]
        length = singleton["norm"].iloc[0]
        norm_lengths.append(length)
    norm_lengths = np.array(norm_lengths)
    norm_lengths = do_cuts(norm_lengths, hi_cutoff, lo_cutoff, "len")
    logger.info(
        "%d singleton outliers in clusters of size %d",
        len(norm_lengths),
        cluster_size,
    )
    logger.info("min:\t%.3f", min(norm_lengths))
    logger.info("maxes:\t%.3f", max(norm_lengths))
    logger.info("mean: %.3f", norm_lengths.mean())
    axis = sns.distplot(norm_lengths, bins=100, kde_kws={"label": "KDE"})
    axis.set_xlabel("Normalized Length of Singleton")
    plt.title(
        "Length distribution of %d singleton subclusters" % (len(norm_lengths))
    )
    outfilename = f"norm_len_dist.{FILETYPE}"
    logger.info("saving plot to %s", outfilename)
    # plt.yscale('log')
    plt.savefig(outfilename, dpi=200)
    plt.show()


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--hi_cutoff",
    default=0.0,
    show_default=True,
    help="Disregard above this value.",
)
@click.option(
    "--lo_cutoff",
    default=0.0,
    show_default=True,
    help="Disregard below this value.",
)
@click.argument("cluster_size")
@click.argument("combinedfile")
def length_std_dist(cluster_size, hi_cutoff, lo_cutoff, combinedfile):
    """Plot length distribution of cluster singletons."""
    cluster_size = int(cluster_size)
    if cluster_size <= 0:
        logger.error("Positive cluster size must be specified")
        sys.exit(1)
    clusters = pd.read_csv(combinedfile, sep="\t", index_col=0)
    stds = []
    for unused_cluster_id, cluster in clusters.groupby(
        "hom.cluster"
    ):  # pylint: disable=unused-variable
        if cluster["siz"].iloc[0] != cluster_size:
            # not the right size
            continue
        if len(set(cluster["sub"])) != 1:
            # Only one subcluster
            continue
        val = cluster["std"].iloc[0]
        stds.append(val)
    stds = np.array(stds)
    pct_zeros = len(stds[stds == 0.0]) * 100 / len(stds)
    stds = do_cuts(stds, hi_cutoff, lo_cutoff, "stds")
    logger.info(
        "%d single-subgroup clusters of size %d", len(stds), cluster_size
    )
    logger.info("%.1f %% zeroes, max is %.2f", pct_zeros, max(stds))
    logger.info("mean is %.3f", stds.mean())
    logbins = np.logspace(0.7, 3, 100)
    axis = sns.distplot(stds, bins=logbins, kde=False)
    axis.set_xlabel("Standard Deviation of Single-Subgroup Clusters")
    title = "Length Standard Deviation distribution of %d clusters" % len(stds)
    plt.title(title)
    outfilename = f"std_dist.{FILETYPE}"
    logger.info("saving plot to %s", outfilename)
    plt.yscale("log")
    plt.xscale("log")
    plt.savefig(outfilename, dpi=200)
    plt.show()


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--hi_cutoff",
    default=1000,
    show_default=True,
    help="Disregard above this value.",
)
@click.option(
    "--lo_cutoff",
    default=1,
    show_default=True,
    help="Disregard below this value.",
)
@click.option(
    "--identity",
    "-i",
    default=0.0,
    help="Minimum sequence ID (0-1). [default: lowest]",
)
@click.argument("setsize", type=click.IntRange(2, 100))
@click.argument("setname")
def plot_degree_dists(identity, hi_cutoff, lo_cutoff, setname, setsize):
    """Plot homology and synteny degree distributions."""
    set_path = Path(setname)
    homo_cluster_name = cluster_set_name(setname, identity)
    homo_degree = pd.read_csv(
        set_path / homo_degree_dist_filename(homo_cluster_name),
        index_col=0,
        sep="\t",
    )
    homo_degree.index.name = "size"
    homo_degree = homo_degree.rename(
        columns={"clusters": "clusts", "pct_total": "%clusts"}
    )
    homo_clusts = sum(homo_degree["clusts"])
    homo_genes = sum(homo_degree["clusts"] * homo_degree.index)

    homo_degree["%genes"] = (
        homo_degree["clusts"] * homo_degree.index * 100.0 / homo_genes
    )
    synteny_degree_path = set_path / "dagchainer" / "clusters-sizedist.tsv"
    synteny_degree = pd.read_csv(synteny_degree_path, index_col=0, sep="\t")
    synteny_clusts = sum(synteny_degree["clusts"])
    synteny_genes = sum(synteny_degree["clusts"] * synteny_degree.index)
    logger.info("  method       clusters  genes")
    logger.info(f"homology:\t{homo_clusts}\t{homo_genes}")
    logger.info(f" synteny:\t{synteny_clusts}\t{synteny_genes}")
    # Make plot
    plt.plot(homo_degree.index, homo_degree["%genes"], label="homology")
    plt.plot(synteny_degree.index, synteny_degree["%genes"], label="synteny")
    plt.style.use("seaborn-whitegrid")
    plt.xlabel("Cluster Size")
    plt.ylabel("% of Genes in Cluster")
    plt.title(
        f"Cluster size distribution of {homo_genes} genes in {setsize}"
        f" {setname} genomes"
    )
    outfilename = f"cluster_size_dist.{FILETYPE}"
    logger.info(f"saving plot to {outfilename}")
    plt.xlim([lo_cutoff, hi_cutoff])
    plt.legend()
    plt.yscale("log")
    plt.xscale("log", basex=setsize)
    plt.savefig(outfilename, dpi=200)
    plt.show()



# -*- coding: utf-8 -*-
"""Send and receive on-disk messages through names pips with file locking."""
# standard library imports
import array
import contextlib
import fcntl
from pathlib import Path

# third-party imports
import attr
import numpy as np
import numpy.ma as ma
import pandas as pd


@attr.s
class DataMailboxes:

    """Pass data to and from on-disk FIFOs."""

    n_boxes = attr.ib()
    mb_dir_path = attr.ib(default=Path("./mailboxes/"))
    file_extension = attr.ib(default=None)

    def write_headers(self, header):
        """Initialize the mailboxes, optionally writing header."""
        self.mb_dir_path.mkdir(parents=True, exist_ok=True)
        for i in range(self.n_boxes):
            mb_path = self.path_to_mailbox(i)
            with mb_path.open("w") as fh:
                fh.write(header)

    @contextlib.contextmanager
    def locked_open_for_write(self, box_no):
        """Open a mailbox with an advisory file lock."""
        mb_path = self.path_to_mailbox(box_no)
        with mb_path.open("a+") as fd:
            fcntl.flock(fd, fcntl.LOCK_EX)
            yield fd
            fcntl.flock(fd, fcntl.LOCK_UN)

    @contextlib.contextmanager
    def open_then_delete(self, box_no, delete=True):
        """Open a mailbox with an advisory file lock."""
        box_path = self.path_to_mailbox(box_no)
        with box_path.open("r") as fd:
            yield fd
            if delete:
                box_path.unlink()

    def path_to_mailbox(self, box_no):
        """Return a path to a mailbox file."""
        if self.file_extension is None:
            ext = ""
        else:
            ext = f".{self.file_extension}"
        return self.mb_dir_path / f"{box_no}{ext}"

    def delete(self):
        """Remove the mailbox directory. """
        file_list = list(self.mb_dir_path.glob("*"))
        for file in file_list:
            file.unlink()
        self.mb_dir_path.rmdir()


@attr.s
class ExternalMerge(object):

    """Merges integers from files."""

    file_path_func = attr.ib(default=None)
    n_merge = attr.ib(default=None)
    value_vec = None
    fh_list = []

    def init(self, header_value):
        """Read and check the header info."""
        self.fh_list = [
            self.file_path_func(i).open() for i in range(self.n_merge)
        ]
        assert [next(fh).rstrip() for fh in self.fh_list] == [
            header_value
        ] * self.n_merge
        self.value_vec = ma.masked_array(
            np.zeros(self.n_merge), mask=np.zeros(self.n_merge),
        ).astype(np.uint64)
        self._next_vals(np.arange(self.n_merge))

    def _next_vals(self, index_vec):
        """Return the value of the next iterated value."""
        for index in index_vec:
            try:
                self.value_vec[index] = int(next(self.fh_list[index]))
            except StopIteration:
                self.value_vec.mask[index] = 1

    def _close_all(self):
        """Close all filehandles."""
        for i in range(self.n_merge):
            self.fh_list[i].close()

    def merge(self, count_key=None, ordinal_key=None):
        """Return list of merged values."""
        hashes = array.array("L")
        counts = array.array("h")
        while (~self.value_vec.mask).sum() > 1:
            minimum = np.amin(self.value_vec)
            where_min = np.where(self.value_vec == minimum)[0]
            self._next_vals(where_min)
            count = len(where_min)
            if count > 1:
                hashes.append(minimum)
                counts.append(count)
        self._close_all()
        merge_frame = pd.DataFrame(
            pd.array(counts, dtype=pd.UInt32Dtype()),
            columns=[count_key],
            index=hashes,
        )
        merge_frame.sort_values(by=[count_key], inplace=True)
        merge_frame["tmp.base"] = pd.array(
            range(len(merge_frame)), dtype=pd.UInt32Dtype()
        )
        return merge_frame



# -*- coding: utf-8 -*-
"""Sequence (FASTA) and genome (GFF) ingestion operations."""
# standard library imports
import contextlib
import json
import os
import shutil
import sys
import tempfile
from pathlib import Path

# third-party imports
import attr
import click
import dask.bag as db
import gffpandas.gffpandas as gffpd
import numpy as np
import pandas as pd
import toml
from dask.diagnostics import ProgressBar

# first-party imports
import smart_open
from loguru import logger
from pathvalidate import ValidationError
from pathvalidate import validate_filename

# module imports
from . import cli
from . import click_loguru
from .common import ALTERNATE_ABBREV
from .common import CHROMOSOME_ABBREV
from .common import CHROMOSOME_SYNONYMS
from .common import DIRECTIONAL_CATEGORY
from .common import FRAGMENTS_FILE
from .common import PLASTID_STARTS
from .common import PROTEINS_FILE
from .common import PROTEOMES_FILE
from .common import SAVED_INPUT_FILE
from .common import SCAFFOLD_ABBREV
from .common import SCAFFOLD_SYNONYMS
from .common import dotpath_to_path
from .common import sort_proteome_frame
from .common import write_tsv_or_parquet
from .core import cleanup_fasta
from .taxonomy import rankname_to_number

# global constants
__all__ = ["TaxonomicInputTable", "read_from_url"]
FILE_TRANSPORT = "file://"
REQUIRED_LEAF_NAMES = (
    "fasta",
    "gff",
)
COMPRESSION_EXTENSIONS = (
    "gz",
    "bz2",
)


@cli.command()
@click_loguru.init_logger()
@click.option(
    "--parallel/--no-parallel",
    is_flag=True,
    default=True,
    show_default=True,
    help="Process in parallel.",
)
@click.argument("input_toml")
def ingest_sequence_data(input_toml, parallel):
    """Marshal protein and genome sequence information.

    IDs must correspond between GFF and FASTA files and must be unique across
    the entire set.
    """
    options = click_loguru.get_global_options()
    input_obj = TaxonomicInputTable(Path(input_toml), write_table=False)
    input_table = input_obj.input_table
    set_path = Path(input_obj.setname)
    arg_list = []
    for i, row in input_table.iterrows():
        arg_list.append((row["path"], row["fasta_url"], row["gff_url"],))
    if parallel:
        bag = db.from_sequence(arg_list)
    else:
        file_stats = []
    if not options.quiet:
        logger.info(f"Extracting FASTA/GFF info for {len(arg_list)} genomes:")
        ProgressBar().register()
    if parallel:
        file_stats = bag.map(read_fasta_and_gff).compute()
    else:
        for args in arg_list:
            file_stats.append(read_fasta_and_gff(args))
    del arg_list
    seq_stats = pd.DataFrame.from_dict([s[0] for s in file_stats]).set_index(
        "path"
    )
    frag_stats = pd.DataFrame.from_dict([s[1] for s in file_stats]).set_index(
        "path"
    )
    proteomes = pd.concat(
        [input_table.set_index("path"), frag_stats, seq_stats], axis=1
    )
    proteomes.drop(["fasta_url", "gff_url"], axis=1, inplace=True)
    proteomes = sort_proteome_frame(proteomes)
    proteome_table_path = set_path / PROTEOMES_FILE
    logger.info(
        f'Writing table of proteomes to "{proteome_table_path}", edit it to'
        + " change preferences"
    )
    write_tsv_or_parquet(proteomes, proteome_table_path)
    idx_start = 0
    for df in [s[2] for s in file_stats]:
        df.index = range(idx_start, idx_start + len(df))
        idx_start += len(df)
    frags = pd.concat([s[2] for s in file_stats], axis=0)
    frags.index.name = "idx"
    fragalyzer = FragmentCharacterizer()
    frags = fragalyzer.assign_frag_properties(frags)
    frags_path = set_path / FRAGMENTS_FILE
    if not frags_path.exists():
        logger.info(f'Writing "{frags_path}", edit it to rename fragments')
        write_tsv_or_parquet(frags, frags_path)
    else:
        new_frags_path = set_path / ("new." + FRAGMENTS_FILE)
        logger.info(f'A fragments file table already exists at "{frags_path}"')
        logger.info(
            f'A new file has been written at "{new_frags_path}"'
            + "edit and rename it to rename fragments"
        )
        write_tsv_or_parquet(frags, new_frags_path)


def read_fasta_and_gff(args):
    """Read corresponding sequence and position files and construct consolidated tables."""
    dotpath, fasta_url, gff_url = args
    out_path = dotpath_to_path(dotpath)

    with read_from_url(fasta_url) as fasta_fh:
        unused_stem, unused_path, prop_frame, file_stats = cleanup_fasta(
            out_path, fasta_fh, dotpath, write_fasta=False, write_stats=False
        )
    # logger.debug(f"Reading GFF file {gff_url}.")
    with filepath_from_url(gff_url) as local_gff_file:
        annotation = gffpd.read_gff3(local_gff_file)
    features = annotation.filter_feature_of_type(
        ["mRNA"]
    ).attributes_to_columns()
    del annotation
    features.drop(
        features.columns.drop(
            ["seq_id", "start", "strand", "ID"]
        ),  # drop EXCEPT these
        axis=1,
        inplace=True,
    )  # drop non-essential columns
    features = features.set_index("ID")
    features = features.rename(
        columns={
            "start": "frag.start",
            "seq_id": "tmp.seq_id",
            "strand": "tmp.strand",
        }
    )
    features.index.name = "prot.id"
    # Make categoricals
    features["frag.id"] = pd.Categorical(features["tmp.seq_id"])
    features["frag.direction"] = pd.Categorical(
        features["tmp.strand"], dtype=DIRECTIONAL_CATEGORY
    )
    # Drop any features not found in sequence file, e.g., zero-length
    features = features[features.index.isin(prop_frame.index)]
    # sort fragments by largest value
    frag_counts = features["frag.id"].value_counts()
    n_frags = len(frag_counts)
    frags = pd.DataFrame(
        {
            "path": [dotpath] * n_frags,
            "frag.len": frag_counts,
            "frag.orig_id": frag_counts.index,
        },
        index=(frag_counts.index),
    )
    frag_stats = {
        "path": dotpath,
        "frag.n": n_frags,
        "frag.max": frag_counts[0],
    }
    features["frag.count"] = features["frag.id"].map(frag_counts)
    features.sort_values(
        by=["frag.count", "frag.start"], ascending=[False, True], inplace=True
    )
    frag_id_range = []
    for frag_id in frag_counts.index:
        frag_id_range += list(range(frag_counts[frag_id]))
    features["frag.pos"] = frag_id_range
    del frag_id_range
    features.drop(["frag.count"], axis=1, inplace=True)
    # join GFF info to FASTA info
    joined_path = out_path / PROTEINS_FILE
    features = features.join(prop_frame)
    write_tsv_or_parquet(features, joined_path)
    return file_stats, frag_stats, frags


@attr.s
class FragmentCharacterizer:

    """Rename and characterize fragments based on those names."""

    extra_plastid_starts = attr.ib(default=[])
    extra_chromosome_starts = attr.ib(default=[])
    extra_scaffold_starts = attr.ib(default=[])

    @attr.s
    class UnknownsAsScaffolds(object):

        """Keep track of unknown scaffolds."""

        scaf_dict = attr.ib(default={})
        n_scafs = attr.ib(default=0)
        extra_plastid_starts = attr.ib(default=[])

        def rename_unknowns(self, string):
            """If the begginning of the id isn't recognized, rename as unique scaffold."""
            for start in (
                [SCAFFOLD_ABBREV, CHROMOSOME_ABBREV]
                + PLASTID_STARTS
                + self.extra_plastid_starts
            ):
                if string.startswith(start):
                    return string
            if string not in self.scaf_dict:
                self.scaf_dict[string] = self.n_scafs
                self.n_scafs += 1
            return f"{ALTERNATE_ABBREV} {self.scaf_dict[string]}"

    @attr.s
    class PreserveUnique(object):

        """Ensure uniqueness of a series is preserved."""

        ser = attr.ib(default=None)

        def ensure_still_unique(self, new_ser):
            if new_ser.nunique() == self.ser.nunique():
                self.ser = new_ser

    def remove_leading_zeroes_in_field(self, string):
        """If blank-separated fields are integer, remove the leading zero."""
        split = string.split(" ")
        for i, field in enumerate(split):
            if field.isdigit():
                split[i] = f"{int(field)}"
        return " ".join(split)

    def is_scaffold(self, ids):
        """Assess whether fragment is likely a scaffold."""
        return np.logical_or(
            ids.str.contains(SCAFFOLD_ABBREV),
            ids.str.startswith(ALTERNATE_ABBREV),
        )

    def is_chromosome(self, ids):
        """Assess whether fragment is likely a chromosome."""
        return np.logical_and(
            ids.str.startswith(CHROMOSOME_ABBREV), ~self.is_scaffold(ids)
        )

    def is_plastid(self, ids):
        """Assess whether fragment is likely a plastid."""
        return np.logical_or.reduce(
            [ids.str.startswith(s) for s in PLASTID_STARTS]
        )

    def cleanup_frag_ids(self, orig_ids):
        """Try to automatically put genome fragments on a common basis."""
        uniq = self.PreserveUnique(ser=orig_ids)
        split_ids = orig_ids.str.split(".", expand=True)
        split_ids = split_ids.drop(
            [i for i in split_ids.columns if len(set(split_ids[i])) == 1],
            axis=1,
        )
        uniq.ensure_still_unique(split_ids.agg(".".join, axis=1))
        uniq.ensure_still_unique(uniq.ser.str.lower())
        uniq.ensure_still_unique(uniq.ser.str.lower().str.replace("_", " "))
        for synonym in CHROMOSOME_SYNONYMS:
            uniq.ensure_still_unique(
                uniq.ser.str.replace(synonym, CHROMOSOME_ABBREV)
            )
        uniq.ensure_still_unique(
            uniq.ser.str.replace(CHROMOSOME_ABBREV, CHROMOSOME_ABBREV + " ")
        )
        for synonym in SCAFFOLD_SYNONYMS:
            uniq.ensure_still_unique(
                uniq.ser.str.replace(synonym, SCAFFOLD_ABBREV)
            )
        uniq.ensure_still_unique(
            uniq.ser.str.replace(SCAFFOLD_ABBREV, SCAFFOLD_ABBREV + " ")
        )
        uniq.ensure_still_unique(uniq.ser.str.replace("  ", " "))
        clean = uniq.ser.map(self.remove_leading_zeroes_in_field)
        sc_namer = self.UnknownsAsScaffolds()
        clean = clean.map(sc_namer.rename_unknowns)
        orig_ids.index = clean
        self.trans_dict = orig_ids.to_dict()
        return clean

    def assign_frag_properties(self, frags):
        """Assign properties to fragments based on cleaned-up names."""
        clean_list = []
        for unused_path, subframe in frags.groupby(by=["path"]):
            clean_list.append(self.cleanup_frag_ids(subframe["frag.orig_id"]))
        clean_ids = pd.concat(clean_list).sort_index()
        frags["frag.is_plas"] = self.is_plastid(clean_ids)
        frags["frag.is_scaf"] = self.is_scaffold(clean_ids)
        frags["frag.is_chr"] = self.is_chromosome(clean_ids)
        frags["frag.id"] = clean_ids
        return frags


class TaxonomicInputTable:

    """Parse an azulejo input dictionary."""

    def __init__(self, toml_path, write_table=True):
        """Create structures."""
        self.depth = 0
        try:
            tree = toml.load(toml_path)
        except TypeError:
            logger.error(f"Error in filename {toml_path}")
            sys.exit(1)
        except toml.TomlDecodeError as e:
            logger.error(f"File {toml_path} is not valid TOML")
            logger.error(e)
            sys.exit(1)
        if len(tree) > 1:
            logger.error(
                f"Input file {toml_path} should define a single "
                + f"object, but defines {len(tree)} instead"
            )
            sys.exit(1)
        self.setname = self._validate_name(list(tree.keys())[0])
        root_path = Path(self.setname)
        if not root_path.exists():
            logger.info(f"Creating directory for set {self.setname}")
            root_path.mkdir(parents=True)
        self._Node = attr.make_class(
            "Node", ["path", "name", "rank", "rank_val"]
        )
        self._nodes = []
        self._genome_dir_dict = {}
        self._n_genomes = 0
        self._walk(self.setname, tree[self.setname])
        self.input_table = pd.DataFrame.from_dict(
            self._genome_dir_dict
        ).transpose()
        del self._genome_dir_dict
        del self._nodes
        self.input_table.index.name = "order"
        if write_table:
            input_table_path = root_path / PROTEOMES_FILE
            logger.debug(
                f"Input table of {len(self.input_table)} genomes written to"
                f" {input_table_path}"
            )
            write_tsv_or_parquet(self.input_table, input_table_path)
        saved_input_path = root_path / SAVED_INPUT_FILE
        if toml_path != saved_input_path:
            shutil.copy2(toml_path, root_path / SAVED_INPUT_FILE)

    def _validate_name(self, name):
        """Check if a potential filename is valid or not."""
        try:
            validate_filename(name)
        except ValidationError as e:
            logger.error(f"Invalid component name {name} in input file")
            sys.exit(1)
        return name

    def _validate_uri(self, uri):
        """Check if the transport at the start of a URI is valid or not."""
        try:
            smart_open.parse_uri(uri)
        except NotImplementedError:
            logger.error(f'Unimplemented transport in uri "{uri}"')
            sys.exit(1)
        return uri

    def _strip_file_uri(self, url):
        """Removes the file:// uri from a URL string."""
        if url.startswith(FILE_TRANSPORT):
            return url[len(FILE_TRANSPORT) :]
        return url

    def _walk(self, node_name, tree):
        """Recursively walk tree structure."""
        # Check for required field properties.
        if len(self._nodes) > 0:
            dot_path = f"{self._nodes[-1].path}.{node_name}"
        else:
            dot_path = node_name
        if "name" not in tree:
            tree["name"] = f"'{node_name}'"
        if "rank" not in tree:
            logger.error(f'Required entry "rank" not in entry {dot_path}')
            sys.exit(1)
        try:
            rank_val = rankname_to_number(tree["rank"])
        except ValueError as e:
            logger.error(f"Unrecognized taxonomic rank {tree['rank']}")
            logger.error(e)
            sys.exit(1)
        if (len(self._nodes) > 0) and rank_val <= self._nodes[-1].rank_val:
            logger.error(
                f"rank {tree['rank']} value {rank_val} is not less than"
                + f" previous rank value of {self._nodes[-1].rank_val}"
            )
            sys.exit(1)
        # Push node onto stack
        self._nodes.append(
            self._Node(
                self._validate_name(dot_path),
                tree["name"],
                tree["rank"],
                rank_val,
            )
        )
        self.depth = max(self.depth, len(self._nodes))
        # Initialize node properties dictionary
        properties = {"path": dot_path, "children": []}
        for k, v in tree.items():
            if isinstance(v, dict):
                properties["children"] += [k]
                self._walk(k, v)
            else:
                properties[k] = v
        if len(properties["children"]) == 0:
            del properties["children"]
        # Check if this is a genome directory node
        genome_dir_properties = [
            (p in properties) for p in REQUIRED_LEAF_NAMES
        ]
        if any(genome_dir_properties):
            if not all(genome_dir_properties):
                missing_properties = [
                    p
                    for i, p in enumerate(REQUIRED_LEAF_NAMES)
                    if not genome_dir_properties[i]
                ]
                logger.error(
                    f"Missing properties {missing_properties} "
                    + f"for node {dot_path}"
                )
                sys.exit(1)
            if "uri" not in tree:
                uri = FILE_TRANSPORT
            else:
                uri = self._validate_uri(tree["uri"])
                if not uri.endswith("/"):
                    uri += "/"
            self._genome_dir_dict[self._n_genomes] = {"path": dot_path}
            if "preference" not in tree:
                self._genome_dir_dict[self._n_genomes]["preference"] = ""
            else:
                self._genome_dir_dict[self._n_genomes]["preference"] = tree[
                    "preference"
                ]
            for n in self._nodes:
                self._genome_dir_dict[self._n_genomes][
                    f"phy.{n.rank}"
                ] = n.name
            self._genome_dir_dict[self._n_genomes][
                "fasta_url"
            ] = self._strip_file_uri(uri + tree["fasta"])
            self._genome_dir_dict[self._n_genomes][
                "gff_url"
            ] = self._strip_file_uri(uri + tree["gff"])
            self._n_genomes += 1
        for n in self._nodes:
            properties[n.rank] = n.name
        node_path = dotpath_to_path(dot_path)
        node_path.mkdir(parents=True, exist_ok=True)
        properties_file = node_path / "node_properties.json"
        logger.debug(f"Writing properties file to {properties_file}")
        with properties_file.open("w") as filepointer:
            json.dump(properties, filepointer)
        # Pop node from stack
        self._nodes.pop()


@contextlib.contextmanager
def _cd(newdir, cleanup=lambda: True):
    "Change directory with cleanup."
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)
        cleanup()


@contextlib.contextmanager
def read_from_url(url):
    """Read from a URL transparently decompressing if compressed."""
    yield smart_open.open(url)


@contextlib.contextmanager
def filepath_from_url(url):
    """Get a local file from a URL, decompressing if needed."""
    filename = url.split("/")[-1]
    compressed = False
    uncompressed_filename = filename
    for ext in COMPRESSION_EXTENSIONS:
        if filename.endswith(ext):
            compressed = True
            uncompressed_filename = filename[: -(len(ext) + 1)]
            break
    if (
        url.find("://") == -1 and not compressed
    ):  # no transport, must be a file
        yield url
    else:
        dirpath = tempfile.mkdtemp()
        filehandle = smart_open.open(url)
        dldata = filehandle.read()

        def cleanup():
            shutil.rmtree(dirpath)

        with _cd(dirpath, cleanup):

            with open(uncompressed_filename, "w") as f:
                f.write(dldata)
            tmpfile = str(Path(dirpath) / uncompressed_filename)
            yield tmpfile



# -*- coding: utf-8 -*-
"""Operations with pyarrow."""
# standard library imports
import functools
import io
import sys
import timeit
from pathlib import Path

# third-party imports
import click
import pandas as pd

# first-party imports
from loguru import logger

# module imports
from . import cli
from . import click_loguru

COMPRESSION_TYPES = [
    "SNAPPY",
    "GZIP",
    "BROTLI",
    "NONE",
    "LZ4",
    "LZO",
    "ZSTD",
    None,
]
N_TIMINGS = 10


def check_compression(compression):
    """Check to see if the specified compression is recognized."""
    if compression not in COMPRESSION_TYPES:
        logger.error(f"Unrecognized compression type {compression}")
        logger.error(f"Valid compressions are: {COMPRESSION_TYPES}")
        sys.exit(1)


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--columns",
    default=False,
    is_flag=True,
    show_default=False,
    help="Print names/dtypes of columns.",
)
@click.option(
    "--index_name",
    default=False,
    is_flag=True,
    show_default=False,
    help="Print the name of the index.",
)
@click.option(
    "--no_index",
    default=False,
    is_flag=True,
    show_default=False,
    help="Do not write index column.",
)
@click.option(
    "--pretty",
    default=False,
    is_flag=True,
    show_default=False,
    help="Pretty-print output.",
)
@click.option(
    "--no_header",
    "-h",
    default=False,
    is_flag=True,
    show_default=False,
    help="Do not write the header.",
)
@click.option(
    "--col",
    "-c",
    default=None,
    multiple=True,
    show_default=True,
    help="Write only the named column.",
)
@click.option(
    "--index_val",
    "-i",
    default=None,
    multiple=True,
    show_default=True,
    help="Write only the row with this index value.",
)
@click.option(
    "--head",
    default=0,
    show_default=False,
    help="Write only the first N rows.",
)
@click.option(
    "--max_rows",
    default=None,
    show_default=False,
    help="Pretty-print N rows.",
)
@click.option(
    "--max_cols",
    default=None,
    show_default=False,
    help="Pretty-print N cols.",
)
@click.option(
    "--tail",
    default=0,
    show_default=False,
    help="Write only the last N rows.",
)
@click.argument("parquetfile", type=click.Path(exists=True))
@click.argument("tsvfile", nargs=-1)
def parquet_to_tsv(
    parquetfile,
    tsvfile,
    columns,
    index_name,
    col,
    no_index,
    head,
    tail,
    index_val,
    no_header,
    pretty,
    max_rows,
    max_cols,
):
    """Reads parquet file, writes tsv."""
    parquetpath = Path(parquetfile)
    write_index = not no_index
    write_header = not no_header
    if max_rows is not None:
        max_rows = int(max_rows)
    if max_cols is not None:
        max_cols = int(max_cols)
    if tsvfile == ():
        tsvfile = sys.stdout
    if not parquetpath.exists():
        logger.error(f'parquet file "{parquetpath} does not exist.')
        sys.exit(1)
    df = pd.read_parquet(parquetpath)
    if index_val != ():
        index_values = []
        for val in index_val:
            if val not in df.index:
                logger.error(f'name "{val}" not found in index')
                sys.exit(1)
            index_values.append(val)
        df = df[df.index.isin(index_values)]
    elif head > 0:
        df = df[:head]
    elif tail > 0:
        df = df[tail:]
    if columns:
        print(df.dtypes)
    elif index_name:
        print(df.index.name)
    else:
        if col != ():
            col_list = []
            for c in col:
                if c not in df.columns:
                    logger.error(f'name "{c}" not found in list of columns ')
                    sys.exit(1)
                col_list.append(col)
            if pretty:
                with pd.option_context(
                    "display.max_rows",
                    max_rows,
                    "display.max_columns",
                    max_cols,
                    "display.float_format",
                    "{:,.2f}%".format,
                ):
                    print(df)
            else:
                df.to_csv(
                    tsvfile,
                    sep="\t",
                    columns=col_list,
                    index=write_index,
                    header=write_header,
                )
        else:
            if pretty:
                with pd.option_context(
                    "display.max_rows",
                    max_rows,
                    "display.max_columns",
                    max_cols,
                    "display.float_format",
                    "{:,.2f}%".format,
                ):
                    print(df)
            else:
                df.to_csv(
                    tsvfile, sep="\t", index=write_index, header=write_header
                )


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("compression")
@click.argument("infile", type=click.Path(exists=True))
def change_compression(infile, compression):
    """Change the compression type of a Parquet file."""
    pd.read_parquet(infile).to_parquet(
        infile, compression=check_compression(compression)
    )


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--compression",
    "-c",
    default=None,
    show_default=True,
    help="Specify the compression to be used.",
)
@click.argument("tsvfile", type=click.Path(exists=True))
@click.argument("parquetfile", type=click.Path())
def tsv_to_parquet(tsvfile, parquetfile, compression):
    """Write a TSV file as a parquet file using specified compression."""
    pd.read_csv(tsvfile, sep="\t", index_col=0).to_parquet(
        parquetfile, compression=check_compression(compression)
    )


def parse_parquet(filebuf):
    """Use pandas to parse a Parquet file."""
    pd.read_parquet(filebuf)


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("infile", type=click.Path(exists=True))
def time_parquet_parsing(infile):
    """Calculate the time it takes to parse a buffered file"""
    filebuf = io.BytesIO(Path(infile).open("rb").read())
    t = timeit.Timer(functools.partial(parse_parquet, filebuf))
    time_s = t.timeit(N_TIMINGS)
    logger.info(f"File parsed in memory in {time_s:.2f} s")



# -*- coding: utf-8 -*-
"""Constants and functions in common across modules."""
# standard library imports
import contextlib
import json
import mmap
import sys
from pathlib import Path

# third-party imports
import pandas as pd
from loguru import logger

# global constants
NAME = "azulejo"
DEFAULT_PARQUET_COMPRESSION = "zstd"
# Sizes and minimum read times with various compressions
# for a file with one proteome on a system with M.2 SSD disk:
#   "NONE": 43MB, 1.3s
#   "ZSTD": 13M, 1.5s
#    "LZ4": 23MB, 1.5s
# "SNAPPY": 29 MB, 1.5s
# "BROTLI": 13 MB, 1.6s
#   "GZIP": 14 MB, 1.8 s
#    "LZO": not supported in version tested
#    'BZ2': not supported in version tested
# In addition, the ingest process took 28.8s with None, and
# 28.4 s with ZSTD.  Although 'NONE' is marginally faster to
# read the whole file, the margin is low.  At 70% compression,
# the performance of ZSTD can be expected to get better
# on production systems with slower disks and systems for which
# the cache is not warmed up as mine was for testing.
# So I choose ZSTD for now.

PARQUET_EXTENSIONS = ["parquet", "pq", "parq"]
TSV_EXTENSIONS = ["tsv"]
SAVED_INPUT_FILE = "input.toml"

# Changing the extension of these files will change the type of file written.
# TSV files, though readable/editable, do not give the written values back.
# Parquet is also ~100X faster.
4
CLUSTER_FILETYPE = "parq"
CLUSTERS_FILE = "clusters.parq"
FRAGMENTS_FILE = "fragments.tsv"
ANCHOR_HIST_FILE = "anchor_hist.tsv"
HOMOLOGY_FILE = "proteins.hom.parq"
PROTEOMES_FILE = "proteomes.tsv"
PROTEOMOLOGY_FILE = "proteomes.hom.parq"
PROTEOSYN_FILE = "proteomes.hom.syn.parq"
PROTEINS_FILE = "proteins.parq"
SYNTENY_FILE = "proteins.hom.syn.parq"

# fragment-name defs
PLASTID_STARTS = ["chromop", "chl", "mt", "mi", "rh", "mu", "le", "pl"]
CHROMOSOME_SYNONYMS = ["chromosome", "chrom", "chro", "gs", "gm"]
ALTERNATE_ABBREV = "alt"
CHROMOSOME_ABBREV = "chr"
SCAFFOLD_SYNONYMS = ["scaffold", "scaf", "sca"]
SCAFFOLD_ABBREV = "sc"

DIRECTIONAL_CATEGORY = pd.CategoricalDtype(categories=["-", "+"])
# shared functions


def cluster_set_name(stem, identity):
    """Get a setname that specifies the %identity value.."""
    if identity == 1.0:
        digits = "10000"
    else:
        digits = f"{identity:.4f}"[2:]
    return f"{stem}-nr-{digits}"


def get_paths_from_file(filepath, must_exist=True):
    """Given a string filepath,, return the resolved path and parent."""
    inpath = Path(filepath).expanduser().resolve()
    if must_exist and not inpath.exists():
        raise FileNotFoundError(filepath)
    dirpath = inpath.parent
    return inpath, dirpath


class TrimmableMemoryMap:
    """A memory-mapped file that can be resized at the end."""

    def __init__(self, filepath, access=mmap.ACCESS_WRITE):
        """Open the memory-mapped file."""
        self.orig_size = None
        self.size = None
        self.map_obj = None
        self.access = access
        self.filehandle = open(filepath, "r+b")

    def trim(self, start, end):
        """Trim the memory map and mark the nex size."""
        self.map_obj.move(start, end, self.orig_size - end)
        self.size -= end - start
        return self.size

    @contextlib.contextmanager
    def map(self):
        """Open a memory-mapped view of filepath."""
        try:
            self.map_obj = mmap.mmap(
                self.filehandle.fileno(), 0, access=self.access
            )
            self.orig_size = self.map_obj.size()
            self.size = self.orig_size
            yield self.map_obj
        finally:
            if self.access == mmap.ACCESS_WRITE:
                self.map_obj.flush()
                self.map_obj.close()
                self.filehandle.truncate(self.size)
                self.filehandle.close()
            else:
                self.map_obj.close()
                self.filehandle.close()


def dotpath_to_path(dotpath):
    "Return a dot-separated pathstring as a path."
    return Path("/".join(dotpath.split(".")))


def fasta_records(filepath):
    """Count the number of records in a FASTA file."""
    count = 0
    next_pos = 0
    angle_bracket = bytes(">", "utf-8")
    memory_map = TrimmableMemoryMap(filepath, access=mmap.ACCESS_READ)
    with memory_map.map() as mm:
        size = memory_map.size
        next_pos = mm.find(angle_bracket, next_pos)
        while next_pos != -1 and next_pos < size:
            count += 1
            next_pos = mm.find(angle_bracket, next_pos + 1)
    return count, size


def protein_file_stats_filename(setname):
    """Return the name of the protein stat file."""
    if setname is None:
        return "protein_files.tsv"
    return f"{setname}-protein_files.tsv"


def protein_properties_filename(filestem):
    """Return the name of the protein properties file."""
    if filestem is None:
        return "proteins.tsv"
    return f"{filestem}-proteins.tsv"


def homo_degree_dist_filename(filestem):
    """Return the name of the homology degree distribution file."""
    return f"{filestem}-degreedist.tsv"


def group_key_filename(members):
    """Return the name of the group key file."""
    return f"groupkeys-{members}.tsv"


def sort_proteome_frame(df):
    """Sort a proteome frame by preference and frag.max and renumber."""
    df = df.copy()
    if df.index.name == "path":
        df["path"] = df.index
    df.sort_values(
        by=["preference", "frag.max"], ascending=[True, False], inplace=True
    )
    df["order"] = range(len(df))
    df.set_index("order", inplace=True)
    return df


def remove_tmp_columns(df):
    """Remove any columns in a data frame that begin with 'tmp.'."""
    drop_cols = [col for col in df.columns if col.startswith("tmp.")]
    if len(drop_cols) != 0:
        return df.drop(drop_cols, axis=1)
    return df


def write_tsv_or_parquet(
    df,
    filepath,
    compression=DEFAULT_PARQUET_COMPRESSION,
    float_format="%.2f",
    desc=None,
    remove_tmp=True,
):
    """Write either a TSV or a parquet file by file extension."""
    filepath = Path(filepath)
    ext = filepath.suffix.lstrip(".")
    if desc is not None:
        file_desc = f"{desc} file"
        logger.debug(f'Writing {file_desc} "{filepath}')
    if remove_tmp:
        df = remove_tmp_columns(df)
    if ext in PARQUET_EXTENSIONS:
        df.to_parquet(filepath, compression=compression)
    elif ext in TSV_EXTENSIONS:
        df.to_csv(filepath, sep="\t", float_format=float_format)
    else:
        logger.error(f"Unrecognized file extension {ext} in {filepath}")
        sys.exit(1)


def read_tsv_or_parquet(filepath):
    """Read either a TSV or a parquet file by file extension."""
    filepath = Path(filepath)
    if not filepath.exists():
        logger.error(f'File "{filepath}" does not exist.')
        sys.exit(1)
    ext = filepath.suffix.lstrip(".")
    if ext in PARQUET_EXTENSIONS:
        return pd.read_parquet(filepath).convert_dtypes()
    elif ext in TSV_EXTENSIONS:
        df = pd.read_csv(filepath, sep="\t", index_col=0)
        return df.convert_dtypes()
    else:
        logger.error(f"Unrecognized file extensions {ext} in {filepath}")
        sys.exit(1)
